/*
Copyright (C) 1996-1997 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
// Quake is a trademark of Id Software, Inc., (c) 1996 Id Software, Inc. All
// rights reserved.


#include "quakedef.h"
#include "SDL_mixer.h"

Mix_Music *mmusic = NULL;

#define MAXIMUM_TRACKS			100

volatile int idx_buf;

typedef struct
{
	int			start;
	int			length;
	qboolean	isData;
} track_info;

typedef struct
{
	qboolean	valid;
	int			leadOutAddress;
	track_info	track[MAXIMUM_TRACKS];
	byte		lowTrack;
	byte		highTrack;
} cd_info;

static cd_info				cd;

extern	cvar_t	bgmvolume;

static qboolean	playing = false;
static qboolean	wasPlaying = false;
static qboolean	initialized = false;
static qboolean	enabled = true;
static qboolean playLooping = false;
static byte 	remap[MAXIMUM_TRACKS];
//static byte		cdrom;
static byte		playTrack;
//static byte		maxTrack;
static float	cdvolume;


static void CDAudio_Eject(void)
{
}


static void CDAudio_CloseDoor(void)
{
}


static int CDAudio_GetAudioDiskInfo(void)
{

  int size,n,err;

  cd.lowTrack=2;
  cd.highTrack=MAXIMUM_TRACKS;

  FILE *tmpfile;
  for (n = cd.lowTrack; n <= cd.highTrack; n++) {
    tmpfile=fopen(va("cd/%d",n),"rb");
    if (!tmpfile) {
      if (n==cd.lowTrack) {
	cd.valid=false;
	return -1;
      } else {
	cd.highTrack=n-1;
	return 0;
      }
    } else {
      cd.track[n].isData = false;
      cd.track[n].length = size;
    }
  }
  
  return 0;
}


void CDAudio_Play(byte track, qboolean looping)
{

  int err;

  if (!initialized || !enabled)
    return;

  track = remap[track];

  if (playing)
    {
      if (playTrack == track)
	return;
      CDAudio_Stop();
    }

  playLooping = looping;

  Con_DPrintf("CDAudio_Play: loading track number %u.\n", track);
  if (track < cd.lowTrack || track > cd.highTrack)
    {
      Con_DPrintf("CDAudio_Play: Bad track number %u.\n", track);
      return;
    }
  
  playTrack = track;


  musicFilePos=0;
  mmusic=Mix_LoadMUS(va("cd/%d",track));
  if (!mmusic) {
    printf("Failed to load track %d\n",track);
    return;
  }
  Mix_PlayMusic(mmusic,-1);
  playing=true;

  /*
  GpPcmRemove((unsigned short*)musicBuffer);
  err=GpFileOpen(va("gp:\\gpmm\\quake\\cd\\%d.wav",track),OPEN_R,&music_fh);
  if (err==SM_OK) {
    block_in=1;
    GpFileClose(music_fh);
    musicFilePos=0;
    memset(musicBuffer,0,32768);

    GpPcmPlay ( (unsigned short*)musicBuffer, 32768, 1 );
    GpPcmLock ( (unsigned short*)musicBuffer, (int*)&idx_buf, (unsigned int*)&musicPos );
    playing=true;
  }
  */
}


void CDAudio_Stop(void)
{
  if (!initialized || !enabled)
    return;

  wasPlaying=playing;
  playing=false;
  Mix_HaltMusic();
  //  GpPcmRemove((unsigned short*)musicBuffer);
}


void CDAudio_Pause(void)
{
  CDAudio_Stop();
}


void CDAudio_Resume(void)
{
  if (!initialized || !enabled)
    return;
  
  if (!cd.valid)
    return;

  if (!wasPlaying)
    return;

  Mix_PlayMusic(mmusic,-1);

  /*  GpPcmPlay ( (unsigned short*)musicBuffer, 32768, 1 );
  GpPcmLock ( (unsigned short*)musicBuffer, (int*)&idx_buf, (unsigned int*)&musicPos );
  */
  playing=true;
}


static void CD_f (void)
{
	char	*command;
	int		ret;
	int		n;
	int		startAddress;

	if (Cmd_Argc() < 2)
		return;

	command = Cmd_Argv (1);

	if (Q_strcasecmp(command, "seek") == 0)
	{
	  musicFilePos=Q_atoi(Cmd_Argv (2));
	  return;
	}

	if (Q_strcasecmp(command, "on") == 0)
	{
		enabled = true;
		return;
	}

	if (Q_strcasecmp(command, "off") == 0)
	{
		if (playing)
			CDAudio_Stop();
		enabled = false;
		return;
	}

	if (Q_strcasecmp(command, "reset") == 0)
	{
		enabled = true;
		if (playing)
			CDAudio_Stop();
		for (n = 0; n < MAXIMUM_TRACKS; n++)
			remap[n] = n;
		//		CDAudio_Reset();
		CDAudio_GetAudioDiskInfo();
		return;
	}

	if (Q_strcasecmp(command, "remap") == 0)
	{
		ret = Cmd_Argc() - 2;
		if (ret <= 0)
		{
			for (n = 1; n < MAXIMUM_TRACKS; n++)
				if (remap[n] != n)
					Con_Printf("  %u -> %u\n", n, remap[n]);
			return;
		}
		for (n = 1; n <= ret; n++)
			remap[n] = Q_atoi(Cmd_Argv (n+1));
		return;
	}

	if (!cd.valid)
	{
		Con_Printf("No CD in player.\n");
		return;
	}

	if (Q_strcasecmp(command, "play") == 0)
	{
		CDAudio_Play(Q_atoi(Cmd_Argv (2)), false);
		return;
	}

	if (Q_strcasecmp(command, "loop") == 0)
	{
		CDAudio_Play(Q_atoi(Cmd_Argv (2)), true);
		return;
	}

	if (Q_strcasecmp(command, "stop") == 0)
	{
		CDAudio_Stop();
		return;
	}

	if (Q_strcasecmp(command, "pause") == 0)
	{
		CDAudio_Pause();
		return;
	}

	if (Q_strcasecmp(command, "resume") == 0)
	{
		CDAudio_Resume();
		return;
	}

	if (Q_strcasecmp(command, "eject") == 0)
	{
		if (playing)
			CDAudio_Stop();
		//		CDAudio_Eject();
		cd.valid = false;
		return;
	}

	if (Q_strcasecmp(command, "info") == 0)
	{
		Con_Printf("%u tracks\n", cd.highTrack - cd.lowTrack + 1);

		if (playing) {
			Con_Printf("Currently %s track %u\n", playLooping ? "looping" : "playing", playTrack);
			Con_Printf("Position %d/%d\n", musicFilePos,cd.track[playTrack].length);
		}
		Con_Printf("Volume is %u\n", (int)(bgmvolume.value * 255.0));

		return;
	}
}


void CDAudio_Update(void)
{
		  /*
  int err,bufferstart;

  if (!initialized || !enabled)
    return;


		if (playing) {

		if ( ((*musicPos-(int)musicBuffer<16384) && (block_in==1)) || ((*musicPos-(int)musicBuffer>16384) && (block_in==0)) ) {

		  err=GpFileOpen(va("gp:\\gpmm\\quake\\cd\\%d.wav",playTrack),OPEN_R,&music_fh);
		  if (err==SM_OK) {

		    bufferstart=16384*block_in;
		    ulong num=0;
		    err=GpFileSeek(music_fh,FROM_BEGIN,musicFilePos,(ulong*)&num);
		    GpFileRead(music_fh,musicBuffer+bufferstart,16384,(ulong*)&num);
		    GpFileClose(music_fh);
		    musicFilePos+=16384;
		    if (musicFilePos>=cd.track[playTrack].length) { // file end reached
		      //		      musicFilePos=0;
		      //		      memset(musicBuffer+(16384*block_in^1),0,16384);
		      playing=false;
		      GpPcmRemove((unsigned short*)musicBuffer);
		      if (playLooping)
			CDAudio_Play(playTrack, true);
		    }
		  }

		  int i=16384;
		  int snd_vol,val;
		  
		  snd_vol=(int)(bgmvolume.value * 256);
		  
		  while (i--) {
		    val = (musicBuffer[bufferstart+i] * snd_vol) >> 8;
		    if (val > 0x7fff)
		      val = 0x7fff;
		    else if (val < (short)0x8000)
		      val = (short)0x8000;
		    musicBuffer[i+(16384*block_in)] = val;
		  }
		  
		  block_in^=1;
		}
		}
		  */
}


int CDAudio_Init(void)
{

  int n;

  if (cls.state == ca_dedicated)
    return -1;
  
  if (COM_CheckParm("-nocdaudio"))
    return -1;
  
  for (n = 0; n < MAXIMUM_TRACKS; n++)
    remap[n] = n;




  initialized = true;


  if (CDAudio_GetAudioDiskInfo())
    {
      Con_Printf("CDAudio_Init: No numbered files in quake/cd.\n");
      enabled = false;
    } else {

      Mix_OpenAudio(22050, AUDIO_S16, 2, 4096);

    }
  
  Cmd_AddCommand ("cd", CD_f);
  
  Con_Printf("CD Audio Initialized\n");


  return 0;
}


void CDAudio_Shutdown(void)
{

}
