/*
Copyright (C) 1996-1997 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
// sys_win.c -- Win32 system interface code

#include "quakedef.h"
#include "winquake.h"
//#include "zlib.h"
//#include "../resource.h"
//#include "../zlib/zlib.h"
//#include <gx.h>
//#include <Aygshell.h>
#include "conproc.h"

#include "kbdio.h"
#include <gpstdio.h>

unsigned long rGLOBALCOUNTER;

void RTCInt(void) __attribute__ ((interrupt ("IRQ")));


#define MINIMUM_WIN_MEMORY		(MINIMUM_MEMORY+1000000)
#define MAXIMUM_WIN_MEMORY		0x1000000

#define CONSOLE_ERROR_TIMEOUT	60.0	// # of seconds to wait on Sys_Error running
					//  dedicated before exiting
#define PAUSE_SLEEP		50	// sleep time on pause or minimization
#define NOT_FOCUS_SLEEP		20	// sleep time when not focus

int oldgpkey;

int		starttime;
int pclk;

volatile int		sys_checksum;
qboolean		isDedicated;
static double		lastcurtime = 0.0;
static double		curtime = 0.0;
static qboolean		sc_return_on_enter = false;
static double		pfreq;
static int		lowshift;
//HANDLE			hinput, houtput;

//static char			*tracking_tag = "Clams & Mooses";

#define	KEYBUF_SIZE	256
static int		keybuf[KEYBUF_SIZE];
static int		keybuf_head=0;
static int		keybuf_tail=0;

void MaskExceptions (void);
void Sys_InitFloatTime (void);
void Sys_PushFPCW_SetHigh (void);
void Sys_PopFPCW (void);

/*
================
Sys_PageIn
================
*/
void Sys_PageIn (void *ptr, int size)
{
	byte	*x;
	int		/*j, */m, n;

// touch all the memory to make sure it's there. The 16-page skip is to
// keep Win 95 from thinking we're trying to page ourselves in (we are
// doing that, of course, but there's no reason we shouldn't)
	x = (byte *)ptr;

	for (n=0 ; n<4 ; n++)
	{
		for (m=0 ; m<(size - 16 * 0x1000) ; m += 4)
		{
			sys_checksum += *(int *)&x[m];
			sys_checksum += *(int *)&x[m + 16 * 0x1000];
		}
	}
}


/*
===============================================================================

FILE IO

===============================================================================
*/

//Dan: This struct is used to support both stdio and gzio file IO.
/*
struct SYS_HANDLE {
	//FILE *f;
 	F_HANDLE f;
	int compressed;
};
*/
#define	MAX_HANDLES		2
//struct SYS_HANDLE sys_handles[MAX_HANDLES];
F_HANDLE sys_handles;

int		findhandle (void)
{
	int		i;

	//	for (i=1 ; i<MAX_HANDLES ; i++)
	//	  if (sys_handles[i].f==NULL) {
	  if (sys_handles==NULL) {
			return 1;
	  }
	  Sys_Error ("out of handles");
	  return -1;
}

/*
================
filelength
================
*/
//int filelength (FILE *f)
/*
int filelength (struct SYS_HANDLE *sh)
{
	int		pos;
	int		end;

	pos = ftell (sh->f);
	fseek (sh->f, 0, SEEK_END);
	end = ftell (sh->f);
	fseek (sh->f, pos, SEEK_SET);

	return end;
}
*/
int Sys_FileOpenRead (char *path, int *hndl)
{
  //	gzFile	*f;
	int		i, retval;
	int		compressed=false;
	int		size;
	int len;
	//	F_HANDLE	f;
	ERR_CODE	err;


	//TODO: make sure a handle was returned
	i = findhandle ();

//	f = fopen(path, "rb");
	GpFileGetSize(path, &size);



	
	//	err = GpFileOpen(path, OPEN_R, &sys_handles[i].f);
	err = GpFileOpen(path, OPEN_R, &sys_handles);

	//if(err != SM_OK)
	if (err!=SM_OK)
	{
		*hndl = -1;
		retval = -1;
		//		sys_handles[i].f=NULL;
		sys_handles=NULL;
	}
	
	/*
	f = gzopen(path, "rb");

	if (!f)
	{
		*hndl = -1;
		retval = -1;
	}
	*/
	else
	{
	  //		sys_handles[i].compressed=compressed;

		*hndl = i;
		retval = size;//filelength(&sys_handles[i]);
	}
	return retval;
}

int Sys_FileOpenWrite (char *path)
{
//	FILE	*f;
//	F_HANDLE f;
	int		i;


	//	i = findhandle ();

//	f = fopen(path, "wb");
//	if(GpFileOpen(path, OPEN_R | OPEN_W, &sys_handles[i].f) != SM_OK)
	if(GpFileOpen(path, OPEN_R | OPEN_W, &sys_handles) != SM_OK)
	  //		if(GpFileCreate(path, NOT_IF_EXIST, &sys_handles[i].f) != SM_OK) {
		if(GpFileCreate(path, NOT_IF_EXIST, &sys_handles) != SM_OK) {
		  //		  sys_handles[i].f = NULL;
		  sys_handles = NULL;
		  return -1;
		}
	//	sys_handles[i].compressed = false;
	
	return i;
}

void Sys_FileClose (int handle)
{
	//fclose (sys_handles[handle].f);
  //  	GpFileClose( sys_handles[handle].f);
  	GpFileClose( sys_handles);
	//gzclose(sys_handles[handle].f);
	//	sys_handles[handle].f = NULL;
	sys_handles = NULL;
}

void Sys_FileSeek (int handle, int position)
{
	int old;
	//gzseek (sys_handles[handle].f, position, SEEK_SET);
	//	GpFileSeek( sys_handles[handle].f, FROM_BEGIN, position, &old);
	GpFileSeek( sys_handles, FROM_BEGIN, position, &old);
}

int Sys_FileRead (int handle, void *dest, int count)
{
	int		x;

	//x = gzread (sys_handles[handle].f,dest,count);
	//	gm_sprintf(gpstr, "h=%i.%i, c=%i", handle, sys_handles[handle].f, count);
	//	GpFileRead(sys_handles[handle].f, dest, count, &x);
	GpFileRead(sys_handles, dest, count, &x);
	return x;
}

int Sys_FileWrite (int handle, void *data, int count)
{
	int		x;
	//x = fwrite (data, 1, count, sys_handles[handle].f);
	//	if(GpFileWrite(sys_handles[handle].f, data, count) != SM_OK)
	if(GpFileWrite(sys_handles, data, count) != SM_OK)
		return 0;
	return count;
}

int	Sys_FileTime (char *path)
{
//	FILE	*f;
	int		retval;

	if( GpFileGetSize(path, &retval) != SM_OK)
		return -1;

	return 1;
/*
//	f = fopen(path, "rb");
	if (f)
	{
		fclose(f);
		retval = 1;
	}
	else
	{
		retval = -1;
	}

	return retval;
*/
}

void Sys_mkdir (char *path)
{
	GpDirCreate(path, NOT_IF_EXIST);
}

void GP_ScanString(F_HANDLE *f,char *string,char splitchar)
{
  char newchar;
  ulong dummy;

  GpFileRead(*f,&newchar,1,&dummy);
  while(newchar != splitchar)
    {
      *string++ = newchar;
      //gzread(f,&newchar,1);
      GpFileRead(*f,&newchar,1,&dummy);
    }
  *string++ = '\0';
}

void GP_ScanFloat(F_HANDLE *f,float *fl)
{
  char sbuf[100];
  GP_ScanString(f, sbuf,'\n');
  *fl=atof(sbuf);
}

GP_ScanInt(F_HANDLE *f,int *in)
{
  char sbuf[100];
  GP_ScanString(f, sbuf,'\n');
  *in=atoi(sbuf);
}

/*
===============================================================================

SYSTEM IO

===============================================================================
*/

/*
================
Sys_MakeCodeWriteable
================
*/
void Sys_MakeCodeWriteable (unsigned long startaddr, unsigned long length)
{
}


//#ifndef _M_IX86

void Sys_SetFPCW (void)
{
}

void Sys_PushFPCW_SetHigh (void)
{
}

void Sys_PopFPCW (void)
{
}

void MaskExceptions (void)
{
}

//#endif

/*
================
Sys_Init
================
*/
void Sys_Init (void)
{
  
	int i;
	int mclk;
	int clkdivset;
	int clkdivider;

	i = COM_CheckParm ("-gpmclk");
	if (i && i < com_argc-1)
	{
	  mclk=atoi(com_argv[i+1]);
	}
	i = COM_CheckParm ("-gpdivset");
	if (i && i < com_argc-1)
	{
	  clkdivset=atoi(com_argv[i+1]);
	}
	i = COM_CheckParm ("-gpdiv");
	if (i && i < com_argc-1)
	{
	  clkdivider=atoi(com_argv[i+1]);
	}

	if (!mclk || !clkdivset || !clkdivider) {
	  mclk=132000000;
	  clkdivset=0x3a011;
	  clkdivider=3;
	}

	// Switch to higher cpu frequency
	GpClockSpeedChange (mclk, clkdivset, clkdivider);
	/*
	if (clkdivider==0) pclk=mclk;
	if (clkdivider==1) pclk=mclk/2;
	if (clkdivider>1) pclk=mclk/4;
	*/
	pclk=GpPClkGet();
	/*	
	for (i=1 ; i<MAX_HANDLES ; i++)
	{
	  sys_handles[i].f=NULL;
	}
	*/
  /*
	MaskExceptions ();
	Sys_SetFPCW ();

	if (!QueryPerformanceFrequency (&PerformanceFreq))
		Sys_Error ("No hardware timer available");

	// get 32 out of the 64 time bits such that we have around
	// 1 microsecond resolution
	lowpart = (unsigned int)PerformanceFreq.LowPart;
	highpart = (unsigned int)PerformanceFreq.HighPart;
	lowshift = 0;

	while (highpart || (lowpart > 2000000.0))
	{
		lowshift++;
		lowpart >>= 1;
		lowpart |= (highpart & 1) << 31;
		highpart >>= 1;
	}

	pfreq = 1.0 / (double)lowpart;

	Sys_InitFloatTime ();

	WinNT=false;
  */
}


void Sys_Error (char *error, ...)
{

	va_list		argptr;
	char		text[1024], text2[1024];
	char		*text3 = "Press Enter to exit\n";
	char		*text4 = "***********************************\n";
	char		*text5 = "\n";
	DWORD		dummy;
	double		starttime;
	static int	in_sys_error0 = 0;
	static int	in_sys_error1 = 0;
	static int	in_sys_error2 = 0;
	static int	in_sys_error3 = 0;

	if (!in_sys_error3)
	{
		in_sys_error3 = 1;
	}

	va_start (argptr, error);
	vsprintf (text, error, argptr);
	va_end (argptr);

/*	if (isDedicated)
	{
		va_start (argptr, error);
		vsprintf (text, error, argptr);
		va_end (argptr);

		sprintf (text2, "ERROR: %s\n", text);
		WriteFile (houtput, text5, strlen (text5), &dummy, NULL);
		WriteFile (houtput, text4, strlen (text4), &dummy, NULL);
		WriteFile (houtput, text2, strlen (text2), &dummy, NULL);
		WriteFile (houtput, text3, strlen (text3), &dummy, NULL);
		WriteFile (houtput, text4, strlen (text4), &dummy, NULL);


		starttime = Sys_FloatTime ();
		sc_return_on_enter = true;	// so Enter will get us out of here

		while (!Sys_ConsoleInput () &&
				((Sys_FloatTime () - starttime) < CONSOLE_ERROR_TIMEOUT))
		{
		}
	}
	else
	{
*/
	GpError(text, 666);

	if (!in_sys_error1)
	{
		in_sys_error1 = 1;
		//		Host_Shutdown ();
	}

// shut down QHOST hooks if necessary
	if (!in_sys_error2)
	{
		in_sys_error2 = 1;
	}

//	exit (1);

	while(1){}
}

void Sys_Printf (char *fmt, ...)
{
/*	va_list		argptr;
	char		text[1024];
	DWORD		dummy;

	if (isDedicated)
	{
		va_start (argptr,fmt);
		vsprintf (text, fmt, argptr);
		va_end (argptr);

		WriteFile(houtput, text, strlen (text), &dummy, NULL);
	}
*/
}

void Sys_Quit (void)
{

	Host_Shutdown();
	GpAppExit();
}


// idea and first written, Jouni Korhonen 
// rewritten and extended (c)2004 Mirko Roller

void RTCInt(void) {
  rGLOBALCOUNTER++;
}


void setupRTCInt( void ) {
   // enable RTC
   rCLKCON |= 0x800;
   //
   // The CPU clock indepentent RTC Timer
   //
   // The ticnt register 0x15700044 (32 bit) offers:
   //
   // BIT   7: 0=disable 1=enable timer
   // BIT 0-6: tick time count value 1-127
   //
   // %10000000 should result in a 1/128 timer, but is not working
   // %10000001 result in 1/64 timer
   // %11111111 result in 1/1  timer
   
   rTICINT = 0x81;
   //(*(volatile unsigned char *)rTICINT) = 0x81; 
   rGLOBALCOUNTER = 0;
}


void InstallRTC( void ) {
     gp_disableIRQ();
     gp_installSWIIRQ(8,RTCInt);
     gp_enableIRQ();   
}

unsigned long gp_getRTC() {
   return rGLOBALCOUNTER;
}

void gp_initRTC() {
      setupRTCInt();
      InstallRTC();
}

void gp_clearRTC() {
   rGLOBALCOUNTER=0;
}
/*
void Sys_GPTIMER(void){
	curtime += 0.005;

}
*/

double temptime = 0;

/*
================
Sys_FloatTime
================
*/
double Sys_FloatTime (void)
{

  /*
	static int			sametimecount;
	static unsigned int	oldtime;
	static int			first = 1;
	LARGE_INTEGER		PerformanceCount;
	unsigned int		temp, t2;
	double				time;

	Sys_PushFPCW_SetHigh ();

	QueryPerformanceCounter (&PerformanceCount);

	temp = ((unsigned int)PerformanceCount.LowPart >> lowshift) |
		   ((unsigned int)PerformanceCount.HighPart << (32 - lowshift));

	if (first)
	{
		oldtime = temp;
		first = 0;
	}
	else
	{
	// check for turnover or backward time
		if ((temp <= oldtime) && ((oldtime - temp) < 0x10000000))
		{
			oldtime = temp;	// so we can't get stuck
		}
		else
		{
			t2 = temp - oldtime;

			time = (double)t2 * pfreq;
			oldtime = temp;

			curtime += time;

			if (curtime == lastcurtime)
			{
				sametimecount++;

				if (sametimecount > 100000)
				{
					curtime += 1.0;
					sametimecount = 0;
				}
			}
			else
			{
				sametimecount = 0;
			}

			lastcurtime = curtime;
		}
	}

	Sys_PopFPCW ();
  */    

  return (float)gp_getRTC()/64;
}


/*
================
Sys_InitFloatTime
================
*/
void Sys_InitFloatTime (void)
{
/*	int		j;

	Sys_FloatTime ();

	j = COM_CheckParm("-starttime");

	if (j)
	{
		curtime = (double) (Q_atof(com_argv[j+1]));
	}
	else
	{
		curtime = 0.0;
	}
*/
	curtime = 0.0;
	lastcurtime = curtime;
}


char *Sys_ConsoleInput (void)
{
	return NULL;
}

void Sys_Sleep (void)
{
//	Sleep (1);
}

extern int vid_paused;





void Sys_TrapKeys()
{
	int key;
	GpKeyGetEx(&key);

	keybuf[keybuf_head] = key;
	keybuf_head = (keybuf_head + 1) & (KEYBUF_SIZE-1);
}

void Sys_SendKeyEvents (void)
{

  int cbch;
  cbch=1;
  while ((cbch!='\r') && (cbch!=0) && ((cbch<32) || (cbch>127))) {
    cbch = pollKbdKey();
    if (cbch=='a') {
      Key_Event(K_LEFTARROW, true);
      Key_Event(K_LEFTARROW, false);
      break;
    }
    if (cbch=='s') { // shift+space
      break;
    }
    if (cbch=='b') {
      Key_Event(K_RIGHTARROW, true);
      Key_Event(K_RIGHTARROW, false);
      break;
    }
    if (cbch=='\r') {
      Key_Event(K_BACKSPACE, true);
      Key_Event(K_BACKSPACE, false);
      break;
    }
    if (cbch=='n') {
      Key_Event(K_ENTER, true);
      Key_Event(K_ENTER, false);
      break;
    }
    if (((cbch>31) && (cbch<128))) {
      if ((cbch>63) && (cbch<91)) {
	cbch+=32;
      }
      Key_Event(cbch, true);
      Key_Event(cbch, false);
      break;
    }
  }


  int gpkey;

//	key = keybuf[keybuf_tail++];
//	keybuf_tail &= (KEYBUF_SIZE-1);

  GpKeyGetEx(&gpkey);
  
  // UP
  if(gpkey & GPC_VK_UP)	Key_Event(K_UPARROW, true);
  else if (oldgpkey & GPC_VK_UP) Key_Event(K_UPARROW, false);
  // DOWN
  if(gpkey & GPC_VK_DOWN)	Key_Event(K_DOWNARROW, true);
  else if (oldgpkey & GPC_VK_DOWN) Key_Event(K_DOWNARROW, false);
  // LEFT
  if(gpkey & GPC_VK_LEFT)	Key_Event(K_LEFTARROW, true);
  else if (oldgpkey & GPC_VK_LEFT) Key_Event(K_LEFTARROW, false);
  // RIGHT
  if(gpkey & GPC_VK_RIGHT)	Key_Event(K_RIGHTARROW, true);
  else if (oldgpkey & GPC_VK_RIGHT)	Key_Event(K_RIGHTARROW, false);
  // ESC
  if(gpkey & GPC_VK_START)	Key_Event(27, true);
  else if (oldgpkey & GPC_VK_START) Key_Event(27, false);
  // CHANGE WEAPON (/)
  if(gpkey & GPC_VK_SELECT)	Key_Event(47, true);
  else if (oldgpkey & GPC_VK_SELECT) Key_Event(47, false);
  
  // MOUSE1 / FIRE
  if(gpkey & GPC_VK_FA)	Key_Event(K_MOUSE1, true);
  else if (oldgpkey & GPC_VK_FA) Key_Event(K_MOUSE1, false);
  // RETURN / JUMP
  if(gpkey & GPC_VK_FB)	Key_Event(K_ENTER, true);
  else if (oldgpkey & GPC_VK_FB) Key_Event(K_ENTER, false);

  /*
  // STRAFE / ALT
  if(gpkey & GPC_VK_FL)	Key_Event(K_ALT, true);
  else if (oldgpkey & GPC_VK_FL) Key_Event(K_ALT, false);
  // LOOK / INS
  if(gpkey & GPC_VK_FR)	Key_Event(K_INS, true);
  else if (oldgpkey & GPC_VK_FR) Key_Event(K_INS, false);
  */

  if(gpkey & GPC_VK_FL && gpkey & GPC_VK_FR){		//default keyboard look key
	  Key_Event(K_INS, true);
	  if(oldgpkey & GPC_VK_FL)
		  Key_Event(',', false);
	  if(oldgpkey & GPC_VK_FR)
		  Key_Event('.', false);
  }
  else if(oldgpkey & GPC_VK_FL && oldgpkey & GPC_VK_FR) Key_Event(K_INS, false);
  else{
	  if(gpkey & GPC_VK_FL)	Key_Event(',', true);
	  else if (oldgpkey & GPC_VK_FL) Key_Event(',', false);
	  if(gpkey & GPC_VK_FR)	Key_Event('.', true);
	  else if (oldgpkey & GPC_VK_FR) Key_Event('.', false);
  }

 

  oldgpkey=gpkey;

}


/*
==============================================================================

 WINDOWS CRAP

==============================================================================
*/


/*
==================
WinMain
==================
*/
void SleepUntilInput (int time)
{
}


/*
==================
WinMain
==================
*/
char		*argv[MAX_NUM_ARGVS];
static char *empty_string = "";
char *lpCmdLine;
char *origlpCmdLine;

double	newtime;

void GpMain(void *gparg)
{
	quakeparms_t	parms;
	double		oldtime, time;
	char	cwd[1024];
	int				t;
	int i;

	doublesize=0;

	GpFatInit();

	lpCmdLine=gm_malloc(500);
	origlpCmdLine=lpCmdLine;

	gm_sprintf(cwd, "gp:\\gpmm\\quake");

	if (cwd[Q_strlen(cwd)-1] == '/')
		cwd[Q_strlen(cwd)-1] = 0;

	parms.basedir = cwd;
	parms.cachedir = NULL;

	parms.argc = 1;
	argv[0] = empty_string;

	sprintf(lpCmdLine,"%s",GpAppPathGet(500));
	if (lpCmdLine[0]=='-') { // launched from gpQuake.fxe or similar

	while (*lpCmdLine && (parms.argc < MAX_NUM_ARGVS))
	{
		while (*lpCmdLine && ((*lpCmdLine <= 32) || (*lpCmdLine > 126)))
			lpCmdLine++;

		if (*lpCmdLine)
		{
			argv[parms.argc] = lpCmdLine;
			parms.argc++;

			while (*lpCmdLine && ((*lpCmdLine > 32) && (*lpCmdLine <= 126)))
				lpCmdLine++;

			if (*lpCmdLine)
			{
				*lpCmdLine = 0;
				lpCmdLine++;
			}
			
		}
	}
	}

	//	argv[parms.argc+1]=empty_string;;
	parms.argv = argv;

	COM_InitArgv (parms.argc, parms.argv);

	parms.argc = com_argc;
	parms.argv = com_argv;

	isDedicated = 0;
	
    	Sys_Init ();

	// force to minimum memory to free up space for gp functions
		parms.memsize = MINIMUM_MEMORY;
		//	parms.memsize = MINIMUM_WIN_MEMORY;

	while (parms.memsize >= MINIMUM_MEMORY) {
	  //		while (parms.memsize >= 0) {
		parms.membase = gm_malloc (parms.memsize);
		if (!parms.membase) parms.memsize-=300000;
		else break;
	}

	/*			
	char tmpstr[20];
	sprintf(tmpstr,"%d",parms.memsize);
	GpError(tmpstr,69);
	*/

	if (!parms.membase)
	  Sys_Error ("Not enough memory free; check disk space\n");

	Sys_Printf ("Host_Init\n");

	Host_Init (&parms);

	GpError("Host_Init", 0);

	int n;
	n = preInitKbd(9600,pclk); // init chatboard - causes error if called before host_init for some reason

  

	// Set up keyboard polling
	//	GpTimerOptSet(0, 100, 0, Sys_GPTIMER);
	//	GpTimerSet(0);
	gp_initRTC();

	//	time = 0.1;
	oldtime = Sys_FloatTime ();

	//	gp32_initrendercache();


    /* main window message loop */

	while (1)
	{
		newtime = Sys_FloatTime ();
		time = newtime - oldtime;
		Host_Frame ((float)time);
		oldtime = newtime;
	}

	//	GpTimerKill(0);

	while (parms.argc--)
		gm_free(argv[parms.argc]);


}

//Dan:
void Sys_HighFPPrecision (void)
{
}

void Sys_LowFPPrecision (void)
{
}
