/************************************************************************
 *									*
 *            MSDOS.H							*
 *            =======							*
 *                                        W.G. Reissnegger		*
 *                                        C&C Research Labs		*
 *                                        NEC USA Inc.			*
 *                                        4 Independence Way		*
 *                                        Princeton NJ 08540		*
 *									*
 *   FAT12 and FAT16 DOS filing systems internal header file            *
 *                                                                      *
 *   Copyright (C) 1997, 1998, 1999, 2000 NEC USA, Inc.			*
 *                 All Rights Reserved					*
 *									*
 *   This program is free software; you can redistribute it and/or      *
 *   modify it under the terms of the GNU General Public License as     *
 *   published by the Free Software Foundation; either version 2 of     *
 *   the license, or (at your option) any later version.                *
 *									*
 *   This program is distributed in the hope that it will be useful,    *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of     *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the      *
 *   GNU General Public License for more details.                       *
 *                                                                      *
 *   You should have received a copy of the GNU General Public          *
 *   License along with this program;  if not, write to the             *
 *       Free Software Foundation, Inc.,                                *
 *       59 Temple Place - Suite 330,                                   *
 *       Boston, MA 02111-1307                                          *
 *                                                                      *
 ************************************************************************/

#ifndef MSDOS_H
#define MSDOS_H

#include <sys/stdtypes.h>
#include <fslib.h>

/*
 *  The partition table for a DOS disk
 */

#define IDE_PARTITION_ACTIVE	0x80

typedef struct
{
    uchar active;
    char start_chs[3];
    uchar type;
    char end_chs[3];
    uint start_lba;
    uint size;
}MSDOS_PARTITION_TABLE;

typedef struct
{
    char ignored[3];			/* Boot strap short or near jump */
    char system_id[8];			/* Name - can be used to special case
						 partition manager volumes */
    uchar sector_size[2];		/* bytes per logical sector */
    uchar cluster_size;			/* sectors/cluster */
    ushort reserved;			/* reserved sectors */
    uchar fats;				/* number of FATs */
    uchar dir_entries[2];		/* root directory entries */
    uchar sectors[2];			/* number of sectors */
    uchar media;			/* media code (unused) */
    ushort fat_length;			/* sectors/FAT */
    ushort secs_track;			/* sectors per track */
    ushort heads;			/* number of heads */
    uint hidden;			/* hidden sectors (unused) */
    uint total_sect;			/* number of sectors (if sectors == 0) */
    uchar drive_id[2];			/* phys drive IDs (0x80, 0x81) */
    uchar ext_boot_sign;		/* extended boot signature */
    uchar vol_serno[2];			/* volume serial number */
    uchar unused[2];
    char volume_name[11];		/* volume name */
    char file_id[8];			/* FAT16 FAT12 */
}MSDOS_BOOT_SECTOR;

#define DOS_FAT16		0x00
#define DOS_FAT12		0x01

#define FAT_EOF			0xffff

#define FAT16_END_MASK		0xfff8
#define FAT16_BAD_BLOCK		0xfff7

#define FAT12_END_MASK		0xff8
#define FAT12_BAD_BLOCK		0xff7

#define FAT16_IS_EOF(_v) 	(((_v) & FAT16_END_MASK) == FAT16_END_MASK)
#define FAT12_IS_EOF(_v) 	(((_v) & FAT12_END_MASK) == FAT12_END_MASK)

typedef struct
{
    uint totalsize;			/* total size in bytes */

    uint bs_lba;
    ushort cluster_size;
    ushort sector_size;
    ushort blksize;

    uchar *fat;
    uint fat_length;
    uint fat_entries;
    uint fat_lba;
    uchar fat_type;			/* DOS_FAT16, DOS_FAT12 */

    ushort root_dir_lba;
    ushort root_dir_entries;

    uint cluster0_lba;
}MSDOS_FS_INFO;

/*
 *  The directory entry
 */

#define ATTR_NONE		0	/* no attribute bits */
#define ATTR_RO			1	/* read-only */
#define ATTR_HIDDEN		2	/* hidden */
#define ATTR_SYS		4	/* system */
#define ATTR_VOLUME		8	/* volume label */
#define ATTR_DIR		16	/* directory */
#define ATTR_ARCH		32	/* archived */

#define ATTR_EXT		(ATTR_RO | ATTR_HIDDEN | ATTR_SYS | ATTR_VOLUME)

#define DELETED_FLAG		0xe5	/* 1st char of deleted file */

typedef struct
{
    char name[8], ext[3];		/* name and extension */
    uchar attr;				/* attribute bits */
    uchar lcase;			/* Case for base and extension */
    uchar ctime_ms;			/* Creation time, milliseconds */
    ushort ctime;			/* Creation time */
    ushort cdate;			/* Creation date */
    ushort adate;			/* Last access date */
    uchar unused[2];
    ushort time, date, start;		/* time, date and first cluster */
    uint size;				/* file size (in bytes) */
}MSDOS_DIR_ENTRY;

typedef struct dos_device_d
{
    struct dos_device_d *next;

    char devid;
    char proc_name[32];

    FILE *fp;
    ROME_PROCESS *proc;

    int inf_valid;
    MSDOS_FS_INFO inf;
}DOS_DEVICE;

typedef struct dos_openfile_d
{
    struct dos_openfile_d *link;

    /* entry of device */
    DOS_DEVICE *dev;

    int type;				/* FS_TYPE_DIR, FS_TYPE_FILE */

    char filename[256];			/* filename of opened file */
    uint mode;				/* open mode as bitfield  */

    int present;			/* buffer present flag */
    int dirty;

    uint fpos;				/* current file position */
    uint fend;				/* current end of file */
    uint flim;				/* limit for current cluster */

    uint mbstart;			/* start charpos of current mblk */
    uint mbend;				/* end charpos of current mblk */

    ROME_MESSAGE mblk;
    ushort cluster;			/* current cluster for the buffer */
    ushort blkidx;			/* relative idx of the cluster */
    ushort start;			/* start cluster */
    ROME_MQUEUE q;
}DOS_OPENFILE;

typedef struct
{
    DOS_OPENFILE *o_root;
    DOS_DEVICE *dev_list;
}DOS_INFO;

/*
 *    Routines for accessing clusters
 *    ===============================
 *
 */

extern void dos_fs_read_cluster(FILE *, MSDOS_FS_INFO *, ushort, 
				ROME_MESSAGE *);

extern void dos_fs_write_dir_cluster(FILE *, MSDOS_FS_INFO *, ushort,
				     ROME_MESSAGE *);

extern void dos_fs_read_dir_cluster(FILE *, MSDOS_FS_INFO *, ushort, 
				    ROME_MESSAGE *);
extern void dos_fs_write_cluster(FILE *, MSDOS_FS_INFO *, ushort,
				 ROME_MESSAGE *);

/*
 *    Routines for accessing the FAT
 *    ==============================
 *
 */

extern int dos_fat_read(FILE *, MSDOS_FS_INFO *);

extern int dos_fat_write(FILE *, MSDOS_FS_INFO *);

extern ushort dos_fat_lookup_entry(ushort, MSDOS_FS_INFO *);

extern int dos_fat_write_entry(MSDOS_FS_INFO *, ushort, ushort);

extern ushort dos_fat_lookup_free(MSDOS_FS_INFO *);

extern ushort dos_fat_seek_entry(MSDOS_FS_INFO *, ushort, register ushort);

extern int dos_fat_free_chain(MSDOS_FS_INFO *, ushort);

/*
 *  Routines for accessing directory entries
 *  ========================================
 */

extern ushort dos_dir_mktime(void);

extern ushort dos_dir_mkdate(void);

extern int dos_dir_entry_exists(FILE *, MSDOS_FS_INFO *, const char *);

extern int dos_dir_dir_entry_exists(FILE *, MSDOS_FS_INFO *, const char *);

extern int dos_dir_get_entry(FILE *, MSDOS_FS_INFO *, const char *,
			     MSDOS_DIR_ENTRY *);

extern int dos_dir_write_entry(FILE *, MSDOS_FS_INFO *, const char *,
			MSDOS_DIR_ENTRY *);

extern int dos_dir_rename_entry(FILE *, MSDOS_FS_INFO *, const char *,
				const char *);

extern int dos_dir_mkfile(FILE *, MSDOS_FS_INFO *, const char *);

extern int dos_dir_mkdir(FILE *, MSDOS_FS_INFO *, const char *);

extern int dos_dir_rmfile(FILE *, MSDOS_FS_INFO *, const char *);

extern int dos_dir_rmdir(FILE *, MSDOS_FS_INFO *, const char *);

extern int dos_dir_read_dir(DOS_OPENFILE *, FSLIB_DIRENT *, int, int *);

/*
 *    Routines for accessing files
 *    ============================
 */

extern int dos_file_exists(FILE *, MSDOS_FS_INFO *, const char *);

extern int dos_file_create(FILE *, MSDOS_FS_INFO *, const char *);

extern int dos_file_close(FILE *, MSDOS_FS_INFO *, const char *, uint);

extern int dos_file_delete(FILE *, MSDOS_FS_INFO *, const char *);

extern int dos_file_get_size(FILE *, MSDOS_FS_INFO *, const char *, uint *);

extern int dos_file_get_start(FILE *, MSDOS_FS_INFO *, const char *, ushort *);

#endif /* MSDOS_H */
