/*
Copyright (C) 1996-1997 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
// sys_win.c -- Win32 system interface code

#include "quakedef.h"
//#include "winquake.h"
#include "gp2xkeydefs.h"

#include <fcntl.h>
#include <sys/resource.h>
#include <sys/mman.h>
//#include <zlib.h>
//#include "../resource.h"
//#include "../zlib/zlib.h"
//#include <gx.h>
//#include <Aygshell.h>
#include "conproc.h"
#include "SDL.h"

static unsigned long mem_dev=0;

       volatile unsigned long  *gp2x_dualcore_ram;


char *basedir = ".";
char *cachedir = "/tmp";

int stickfd=0;
int keyboardfd=0;

int oldgpkey;

#define MINIMUM_WIN_MEMORY		0x0880000
#define MAXIMUM_WIN_MEMORY		0x1000000

#define CONSOLE_ERROR_TIMEOUT	60.0	// # of seconds to wait on Sys_Error running
					//  dedicated before exiting
#define PAUSE_SLEEP		50	// sleep time on pause or minimization
#define NOT_FOCUS_SLEEP		20	// sleep time when not focus


int		starttime;

volatile int		sys_checksum;
qboolean		isDedicated;
static double		lastcurtime = 0.0;
static double		curtime = 0.0;
static qboolean		sc_return_on_enter = false;
static double		pfreq;
static int		lowshift;
//HANDLE			hinput, houtput;

//static char			*tracking_tag = "Clams & Mooses";

#define	KEYBUF_SIZE	256
static int		keybuf[KEYBUF_SIZE];
static int		keybuf_head=0;
static int		keybuf_tail=0;

void MaskExceptions (void);
void Sys_InitFloatTime (void);
void Sys_PushFPCW_SetHigh (void);
void Sys_PopFPCW (void);

/*
================
Sys_PageIn
================
*/
void Sys_PageIn (void *ptr, int size)
{
	byte	*x;
	int		/*j, */m, n;

// touch all the memory to make sure it's there. The 16-page skip is to
// keep Win 95 from thinking we're trying to page ourselves in (we are
// doing that, of course, but there's no reason we shouldn't)
	x = (byte *)ptr;

	for (n=0 ; n<4 ; n++)
	{
		for (m=0 ; m<(size - 16 * 0x1000) ; m += 4)
		{
			sys_checksum += *(int *)&x[m];
			sys_checksum += *(int *)&x[m + 16 * 0x1000];
		}
	}
}


/*
===============================================================================

FILE IO

===============================================================================
*/

//Dan: This struct is used to support both stdio and gzio file IO.

struct SYS_HANDLE {
	FILE *f;
	int compressed;
};

#define	MAX_HANDLES		10
struct SYS_HANDLE sys_handles[MAX_HANDLES];

int		findhandle (void)
{
	int		i;

	for (i=1 ; i<MAX_HANDLES ; i++)
	  if (sys_handles[i].f==NULL) {
	    return i;
	  }
	  Sys_Error ("out of handles");
	  return -1;
}

/*
================
filelength
================
*/
//int filelength (FILE *f)
int Qfilelength (struct SYS_HANDLE *sh)
{
  int             pos;
  int             end;
  /*
  if (sh->compressed) {
    pos = gztell((gzFile)sh->f);
    gzseek ((gzFile)sh->f, 0, SEEK_END);
    end = gztell ((gzFile)sh->f);
    gzseek ((gzFile)sh->f, pos, SEEK_SET);
  } else {
  */
    pos = ftell (sh->f);
    fseek (sh->f, 0, SEEK_END);
    end = ftell (sh->f);
    fseek (sh->f, pos, SEEK_SET);
    //  }

  return end;
}

int Sys_FileOpenRead (char *path, int *hndl)
{
	FILE	*f;
	int		i, retval;
	int		len;
	int		compressed=false;


	//TODO: make sure a handle was returned
	i = findhandle ();
	/*
	len=strlen(path);
	if (len>3) {
		len--;
		if (((path[len]=='z')||(path[len]=='Z'))&&
			((path[len-1]=='g')||(path[len-1]=='G'))&&
			((path[len-2]=='.')||(path[len-2]  =='.')))
			compressed=true;
	}

	if (compressed)
		f = (FILE *)gzopen(path, "rb");
	else
	*/
		f = fopen(path, "rb");

	if (!f)
	{
		*hndl = -1;
		retval = -1;
	}
	else
	{
		sys_handles[i].f = f;
		//		sys_handles[i].compressed=compressed;

		*hndl = i;
		retval = Qfilelength(&sys_handles[i]);
	}

	return retval;
}

int Sys_FileOpenWrite (char *path)
{
	FILE	*f;
	int		i;


	i = findhandle ();

	f = fopen(path, "wb");

	if (!f)
		Sys_Error ("Error opening %s", path);
	sys_handles[i].f = f;
	sys_handles[i].compressed = false;
	

	return i;
}

void Sys_FileClose (int handle)
{
  /*
	if (sys_handles[handle].compressed)
		gzclose(sys_handles[handle].f);
	else
*/
  if (handle>=0) {
    fclose (sys_handles[handle].f);
    sys_handles[handle].f = NULL;
  }
}

void Sys_FileSeek (int handle, int position)
{
  /*
	if (sys_handles[handle].compressed) {
		int curpos=gztell(sys_handles[handle].f);
		gzseek(sys_handles[handle].f, position-curpos, SEEK_CUR);
		//gzseek(sys_handles[handle].f, position, SEEK_SET);
	} else
  */
  if (handle>=0) {
    fseek (sys_handles[handle].f, position, SEEK_SET);
  }
}

int Sys_FileRead (int handle, void *dst, int count)
{
	char *data;
	int size, done;

	size = 0;
	if ( handle >= 0 ) {
		data = dst;
		while ( count > 0 ) {
			done = fread (data, 1, count, sys_handles[handle].f);
			if ( done == 0 ) {
				break;
			}
			data += done;
			count -= done;
			size += done;
		}
	}
	return size;

	/*
	int		x;
	if (sys_handles[handle].compressed)
		x = gzread((gzFile) sys_handles[handle].f, dest, count);
	else

	x = fread (dest, 1, count, sys_handles[handle].f);
	return x;
*/
}

int Sys_FileWrite (int handle, void *src, int count)
{
	char *data;
	int size, done;

	size = 0;
	if ( handle >= 0 ) {
		data = src;
		while ( count > 0 ) {
			done = fread (data, 1, count, sys_handles[handle].f);
			if ( done == 0 ) {
				break;
			}
			data += done;
			count -= done;
			size += done;
		}
	}
	return size;
	/*
	int		x;
	if (!sys_handles[handle].compressed)

	x = fwrite (data, 1, count, sys_handles[handle].f);


//	else //TODO: Support compression writing
	//	return 0;

	return x;
	*/
}

int	Sys_FileTime (char *path)
{
	FILE	*f;
	int		retval;

	f = fopen(path, "rb");

	if (f)
	{
		fclose(f);
		retval = 1;
	}
	else
	{
		retval = -1;
	}
	
	return retval;
}

void Sys_mkdir (char *path)
{
	mkdir(path,777);
}

/*
===============================================================================

SYSTEM IO

===============================================================================
*/

/*
================
Sys_MakeCodeWriteable
================
*/
void Sys_MakeCodeWriteable (unsigned long startaddr, unsigned long length)
{
}


//#ifndef _M_IX86

void Sys_SetFPCW (void)
{
}

void Sys_PushFPCW_SetHigh (void)
{
}

void Sys_PopFPCW (void)
{
}

void MaskExceptions (void)
{
}

//#endif

/*
================
Sys_Init
================
*/
void Sys_Init (void)
{
  
	int i;

	/*
	int mclk;
	int clkdivset;
	int clkdivider;

	i = COM_CheckParm ("-gpmclk");
	if (i && i < com_argc-1)
	{
	  mclk=atoi(com_argv[i+1]);
	}
	i = COM_CheckParm ("-gpdivset");
	if (i && i < com_argc-1)
	{
	  clkdivset=atoi(com_argv[i+1]);
	}
	i = COM_CheckParm ("-gpdiv");
	if (i && i < com_argc-1)
	{
	  clkdivider=atoi(com_argv[i+1]);
	}

	if (!mclk || !clkdivset || !clkdivider) {
	  mclk=132000000;
	  clkdivset=0x3a011;
	  clkdivider=3;
	}

	// Switch to higher cpu frequency
	GpClockSpeedChange (mclk, clkdivset, clkdivider);

	pclk=GpPClkGet();
*/
		
	for (i=1 ; i<MAX_HANDLES ; i++)
	{
	  sys_handles[i].f=NULL;
	}
	
  /*
	MaskExceptions ();
	Sys_SetFPCW ();

	if (!QueryPerformanceFrequency (&PerformanceFreq))
		Sys_Error ("No hardware timer available");

	// get 32 out of the 64 time bits such that we have around
	// 1 microsecond resolution
	lowpart = (unsigned int)PerformanceFreq.LowPart;
	highpart = (unsigned int)PerformanceFreq.HighPart;
	lowshift = 0;

	while (highpart || (lowpart > 2000000.0))
	{
		lowshift++;
		lowpart >>= 1;
		lowpart |= (highpart & 1) << 31;
		highpart >>= 1;
	}

	pfreq = 1.0 / (double)lowpart;

	Sys_InitFloatTime ();

	WinNT=false;
  */
}


void Sys_Error (char *error, ...)
{
    va_list     argptr;
    char        string[1024];

    va_start (argptr,error);
    vsprintf (string,error,argptr);
    va_end (argptr);
	fprintf(stderr, "Error: %s\n", string);

	Host_Shutdown ();
	exit (1);

  /*
	va_list		argptr;
	char		text[1024], text2[1024];
	char		*text3 = "Press Enter to exit\n";
	char		*text4 = "***********************************\n";
	char		*text5 = "\n";
	DWORD		dummy;
	double		starttime;
	static int	in_sys_error0 = 0;
	static int	in_sys_error1 = 0;
	static int	in_sys_error2 = 0;
	static int	in_sys_error3 = 0;

	if (!in_sys_error3)
	{
		in_sys_error3 = 1;
	}

	va_start (argptr, error);
	vsprintf (text, error, argptr);
	va_end (argptr);

//	GpError(text, 666);
	printf(text);
	if (!in_sys_error1)
	{
		in_sys_error1 = 1;
		//		Host_Shutdown ();
	}

// shut down QHOST hooks if necessary
	if (!in_sys_error2)
	{
		in_sys_error2 = 1;
	}

	exit (1);

	//	while(1){}
	*/
}

void Sys_Printf (char *fmt, ...)
{
	va_list		argptr;
	char		text[1024];
	DWORD		dummy;

		va_start (argptr,fmt);
		vsprintf (text, fmt, argptr);
		va_end (argptr);
		printf(text);
	/*
	if (isDedicated)
	{
		va_start (argptr,fmt);
		vsprintf (text, fmt, argptr);
		va_end (argptr);

		WriteFile(houtput, text, strlen (text), &dummy, NULL);
	}
	*/
}

void Sys_Quit (void)
{
	Host_Shutdown();
	SDL_Quit();
}

double temptime = 0;

/*
================
Sys_FloatTime
================
*/
double Sys_FloatTime (void)
{

  /*
	static int			sametimecount;
	static unsigned int	oldtime;
	static int			first = 1;
	LARGE_INTEGER		PerformanceCount;
	unsigned int		temp, t2;
	double				time;

	Sys_PushFPCW_SetHigh ();

	QueryPerformanceCounter (&PerformanceCount);

	temp = ((unsigned int)PerformanceCount.LowPart >> lowshift) |
		   ((unsigned int)PerformanceCount.HighPart << (32 - lowshift));

	if (first)
	{
		oldtime = temp;
		first = 0;
	}
	else
	{
	// check for turnover or backward time
		if ((temp <= oldtime) && ((oldtime - temp) < 0x10000000))
		{
			oldtime = temp;	// so we can't get stuck
		}
		else
		{
			t2 = temp - oldtime;

			time = (double)t2 * pfreq;
			oldtime = temp;

			curtime += time;

			if (curtime == lastcurtime)
			{
				sametimecount++;

				if (sametimecount > 100000)
				{
					curtime += 1.0;
					sametimecount = 0;
				}
			}
			else
			{
				sametimecount = 0;
			}

			lastcurtime = curtime;
		}
	}

	Sys_PopFPCW ();
  */    

    struct timeval tp;
    static int      secbase; 
    
    gettimeofday(&tp, NULL);  

    if (!secbase)
    {
        secbase = tp.tv_sec;
        return tp.tv_usec/1000000.0;
    }

    return (tp.tv_sec - secbase) + tp.tv_usec/1000000.0;

}


/*
================
Sys_InitFloatTime
================
*/
void Sys_InitFloatTime (void)
{
/*	int		j;

	Sys_FloatTime ();

	j = COM_CheckParm("-starttime");

	if (j)
	{
		curtime = (double) (Q_atof(com_argv[j+1]));
	}
	else
	{
		curtime = 0.0;
	}
*/
	curtime = 0.0;
	lastcurtime = curtime;
}


char *Sys_ConsoleInput (void)
{
	return NULL;
}

void Sys_Sleep (void)
{
  SDL_Delay(1);
}

extern int vid_paused;





void Sys_TrapKeys()
{
	int key;
	//	GpKeyGetEx(&key);

	keybuf[keybuf_head] = key;
	keybuf_head = (keybuf_head + 1) & (KEYBUF_SIZE-1);
}


#define	PEPC_VK_UP				(1<<0)	
#define	PEPC_VK_UP_LEFT			(1<<1)	
#define	PEPC_VK_LEFT		(1<<2)	
#define	PEPC_VK_DOWN_LEFT			(1<<3)	
#define	PEPC_VK_DOWN		(1<<4)	
#define	PEPC_VK_DOWN_RIGHT		(1<<5)	
#define	PEPC_VK_RIGHT			(1<<6)	
#define	PEPC_VK_UP_RIGHT			(1<<7)

#define	PEPC_VK_FA				(1<<12)	
#define	PEPC_VK_FB				(1<<13)	
#define	PEPC_VK_FX				(1<<15)	
#define	PEPC_VK_FY				(1<<14)	
#define	PEPC_VK_FL				(1<<10)
#define	PEPC_VK_FR				(1<<11)
#define	PEPC_VK_SELECT				(1<<9)	
#define	PEPC_VK_START				(1<<8)	
#define	PEPC_VK_VOL_UP				(1<<16)	
#define	PEPC_VK_VOL_DOWN	       		(1<<17)	
#define PEPC_VK_STICK                           (1<<18)

unsigned char keymap[128] =
{
    0,  27, '1', '2', '3', '4', '5', '6', '7', '8',	/* 9 */
  '9', '0', '-', '=', K_BACKSPACE,	/* Backspace */
  K_TAB,			/* Tab */
  'q', 'w', 'e', 'r',	/* 19 */
  't', 'y', 'u', 'i', 'o', 'p', '[', ']', K_ENTER,	/* Enter key */
    K_CTRL,			/* 29   - Control */
  'a', 's', 'd', 'f', 'g', 'h', 'j', 'k', 'l', ';',	/* 39 */
 '\'', '`',   K_SHIFT,		/* Left shift */
 '\\', 'z', 'x', 'c', 'v', 'b', 'n',			/* 49 */
  'm', ',', '.', '/',   K_SHIFT,				/* Right shift */
  '*',
    K_ALT,	/* Alt */
  K_SPACE,	/* Space bar */
    0,	/* Caps lock */
    K_F1,	/* 59 - F1 key ... > */
    K_F2,   K_F3,   K_F4,   K_F5,   K_F6,   K_F7,   K_F8,   K_F9,
    K_F10,	/* < ... F10 */
    0,	/* 69 - Num lock*/
    0,	/* Scroll Lock */
    K_HOME,	/* Home key */
    K_UPARROW,	/* Up Arrow */
    K_PGUP,	/* Page Up */
  '-',
    K_LEFTARROW,	/* Left Arrow */
    0,
    K_RIGHTARROW,	/* Right Arrow */
  '+',
    K_END,	/* 79 - End key*/
    K_DOWNARROW,	/* Down Arrow */
    K_PGDN,	/* Page Down */
    K_INS,	/* Insert Key */
    K_DEL,	/* Delete Key */
    0,   0,   0,
    K_F11,	/* F11 Key */
    K_F12,	/* F12 Key */
    0,	/* All other keys are undefined */
};		

void Sys_SendKeyEvents(void)
{
    SDL_Event event;
    int sym, state;
     int modstate;

    while (SDL_PollEvent(&event))
    {
        switch (event.type) {

            case SDL_KEYDOWN:
            case SDL_KEYUP:
                sym = event.key.keysym.sym;
                state = event.key.state;
                modstate = SDL_GetModState();
                switch(sym)
                {
                   case SDLK_DELETE: sym = K_DEL; break;
                   case SDLK_BACKSPACE: sym = K_BACKSPACE; break;
                   case SDLK_F1: sym = K_F1; break;
                   case SDLK_F2: sym = K_F2; break;
                   case SDLK_F3: sym = K_F3; break;
                   case SDLK_F4: sym = K_F4; break;
                   case SDLK_F5: sym = K_F5; break;
                   case SDLK_F6: sym = K_F6; break;
                   case SDLK_F7: sym = K_F7; break;
                   case SDLK_F8: sym = K_F8; break;
                   case SDLK_F9: sym = K_F9; break;
                   case SDLK_F10: sym = K_F10; break;
                   case SDLK_F11: sym = K_F11; break;
                   case SDLK_F12: sym = K_F12; break;
                   case SDLK_BREAK:
                   case SDLK_PAUSE: sym = K_PAUSE; break;
                   case SDLK_UP: sym = K_UPARROW; break;
                   case SDLK_DOWN: sym = K_DOWNARROW; break;
                   case SDLK_RIGHT: sym = K_RIGHTARROW; break;
                   case SDLK_LEFT: sym = K_LEFTARROW; break;
                   case SDLK_INSERT: sym = K_INS; break;
                   case SDLK_HOME: sym = K_HOME; break;
                   case SDLK_END: sym = K_END; break;
                   case SDLK_PAGEUP: sym = K_PGUP; break;
                   case SDLK_PAGEDOWN: sym = K_PGDN; break;
                   case SDLK_RSHIFT:
                   case SDLK_LSHIFT: sym = K_SHIFT; break;
                   case SDLK_RCTRL:
                   case SDLK_LCTRL: sym = K_CTRL; break;
                   case SDLK_RALT:
                   case SDLK_LALT: sym = K_ALT; break;
                   case SDLK_KP0: 
                       if(modstate & KMOD_NUM) sym = K_INS; 
                       else sym = SDLK_0;
                       break;
                   case SDLK_KP1:
                       if(modstate & KMOD_NUM) sym = K_END;
                       else sym = SDLK_1;
                       break;
                   case SDLK_KP2:
                       if(modstate & KMOD_NUM) sym = K_DOWNARROW;
                       else sym = SDLK_2;
                       break;
                   case SDLK_KP3:
                       if(modstate & KMOD_NUM) sym = K_PGDN;
                       else sym = SDLK_3;
                       break;
                   case SDLK_KP4:
                       if(modstate & KMOD_NUM) sym = K_LEFTARROW;
                       else sym = SDLK_4;
                       break;
                   case SDLK_KP5: sym = SDLK_5; break;
                   case SDLK_KP6:
                       if(modstate & KMOD_NUM) sym = K_RIGHTARROW;
                       else sym = SDLK_6;
                       break;
                   case SDLK_KP7:
                       if(modstate & KMOD_NUM) sym = K_HOME;
                       else sym = SDLK_7;
                       break;
                   case SDLK_KP8:
                       if(modstate & KMOD_NUM) sym = K_UPARROW;
                       else sym = SDLK_8;
                       break;
                   case SDLK_KP9:
                       if(modstate & KMOD_NUM) sym = K_PGUP;
                       else sym = SDLK_9;
                       break;
                   case SDLK_KP_PERIOD:
                       if(modstate & KMOD_NUM) sym = K_DEL;
                       else sym = SDLK_PERIOD;
                       break;
                   case SDLK_KP_DIVIDE: sym = SDLK_SLASH; break;
                   case SDLK_KP_MULTIPLY: sym = SDLK_ASTERISK; break;
                   case SDLK_KP_MINUS: sym = SDLK_MINUS; break;
                   case SDLK_KP_PLUS: sym = SDLK_PLUS; break;
                   case SDLK_KP_ENTER: sym = SDLK_RETURN; break;
                   case SDLK_KP_EQUALS: sym = SDLK_EQUALS; break;
                }
                // If we're not directly handled and still above 255
                // just force it to 0
                if(sym > 255) sym = 0;
                Key_Event(sym, state);
                break;

            case SDL_MOUSEMOTION:
                if ( (event.motion.x != (vid.width/2)) ||
                     (event.motion.y != (vid.height/2)) ) {

		    // pipeline mods - fix glitchy touchscreen ---------
		    // Con_Printf ("XRel X  %d   YRel %d \n", event.motion.xrel, event.motion.yrel);

		    // the movement is based on movements relative to last position...
		    // i choose to ignore changes greater than 20... this could probably be upped to 60 or so support
		    // more accelleration in touchscreen
		    if (event.motion.xrel < 20 && event.motion.xrel > -20 && event.motion.yrel < 20 && event.motion.yrel > -20) {
                       mouse_x = event.motion.xrel*10;
                       mouse_y = event.motion.yrel*10;
		    }
	
                    //if ( (event.motion.x < ((vid.width/2)-(vid.width/4))) ||
                    //       (event.motion.x > ((vid.width/2)+(vid.width/4))) ||
                    //        (event.motion.y < ((vid.height/2)-(vid.height/4))) ||
                    //        (event.motion.y > ((vid.height/2)+(vid.height/4))) ) {
 		    //                  SDL_WarpMouse(vid.width/2, vid.height/2);
	            //      }

		    // pipeline end mod --------------------------------
                }
                break;

            case SDL_QUIT:
                CL_Disconnect ();
                Host_ShutdownServer(false);        
                Sys_Quit ();
                break;
            default:
                break;
        }
    }
}
/*
void Sys_SendKeyEvents (void)
{

  int gpkey;

//	key = keybuf[keybuf_tail++];
//	keybuf_tail &= (KEYBUF_SIZE-1);

//  GpKeyGetEx(&gpkey);
  
  read(stickfd, &gpkey, 4); 

  if ((gpkey&PEPC_VK_UP_LEFT) && !((gpkey&PEPC_VK_UP) || (gpkey&PEPC_VK_LEFT))) {
    gpkey=gpkey|PEPC_VK_UP;
    gpkey=gpkey|PEPC_VK_LEFT;
  }
  if ((gpkey&PEPC_VK_UP_RIGHT) && !((gpkey&PEPC_VK_UP) || (gpkey&PEPC_VK_RIGHT))) {
    gpkey=gpkey|PEPC_VK_UP;
    gpkey=gpkey|PEPC_VK_RIGHT;
  }
  if ((gpkey&PEPC_VK_DOWN_LEFT) && !((gpkey&PEPC_VK_DOWN) || (gpkey&PEPC_VK_LEFT))) {
    gpkey=gpkey|PEPC_VK_DOWN;
    gpkey=gpkey|PEPC_VK_LEFT;
  }
  if ((gpkey&PEPC_VK_DOWN_RIGHT) && !((gpkey&PEPC_VK_DOWN) || (gpkey&PEPC_VK_RIGHT))) {
    gpkey=gpkey|PEPC_VK_DOWN;
    gpkey=gpkey|PEPC_VK_RIGHT;
  }
  
  // UP
  if(gpkey & PEPC_VK_UP)	Key_Event(K_UPARROW, true);
  else if (oldgpkey & PEPC_VK_UP) Key_Event(K_UPARROW, false);
  // DOWN
  if(gpkey & PEPC_VK_DOWN)	Key_Event(K_DOWNARROW, true);
  else if (oldgpkey & PEPC_VK_DOWN) Key_Event(K_DOWNARROW, false);
  // LEFT
  if(gpkey & PEPC_VK_LEFT)	Key_Event(K_LEFTARROW, true);
  else if (oldgpkey & PEPC_VK_LEFT) Key_Event(K_LEFTARROW, false);
  // RIGHT
  if(gpkey & PEPC_VK_RIGHT)	Key_Event(K_RIGHTARROW, true);
  else if (oldgpkey & PEPC_VK_RIGHT)	Key_Event(K_RIGHTARROW, false);

  // ESC
  if(gpkey & PEPC_VK_START)	Key_Event(27, true);
  else if (oldgpkey & PEPC_VK_START) Key_Event(27, false);

  // CHANGE WEAPON (/)
  if(gpkey & PEPC_VK_FX)	Key_Event(47, true);
  else if (oldgpkey & PEPC_VK_FX) Key_Event(47, false);
  // RUN
  if(gpkey & PEPC_VK_FY)	Key_Event(K_SHIFT, true);
  else if (oldgpkey & PEPC_VK_FY) Key_Event(K_SHIFT, false);

  // CTRL / FIRE
  if(gpkey & PEPC_VK_FA)	Key_Event(K_CTRL, true);
  else if (oldgpkey & PEPC_VK_FA) Key_Event(K_CTRL, false);
  // RETURN / JUMP
  if(gpkey & PEPC_VK_FB)	Key_Event(K_ENTER, true);
  else if (oldgpkey & PEPC_VK_FB) Key_Event(K_ENTER, false);

  // MOUSE1 / FIRE / MENU SELECT
  if(gpkey & PEPC_VK_STICK)	Key_Event(K_MOUSE1, true);
  else if (oldgpkey & PEPC_VK_STICK) Key_Event(K_MOUSE1, false);

  // something
  if(gpkey & PEPC_VK_SELECT)	Key_Event('b', true);
  else if (oldgpkey & PEPC_VK_SELECT) Key_Event('b', false);

  
  // STRAFE / ALT
  if(gpkey & PEPC_VK_FL)	Key_Event(K_ALT, true);
  else if (oldgpkey & PEPC_VK_FL) Key_Event(K_ALT, false);
  // LOOK / INS
  if(gpkey & PEPC_VK_FR)	Key_Event(K_INS, true);
  else if (oldgpkey & PEPC_VK_FR) Key_Event(K_INS, false);
  

  if(gpkey & PEPC_VK_VOL_UP && gpkey & PEPC_VK_VOL_DOWN){
    Key_Event('a', true); // something
    if(oldgpkey & PEPC_VK_VOL_UP)
      Key_Event('d', false);
    if(oldgpkey & PEPC_VK_VOL_DOWN)
      Key_Event('c', false);
  }
  else if(oldgpkey & PEPC_VK_VOL_DOWN && oldgpkey & PEPC_VK_VOL_UP) Key_Event('a', false);
  else{
    if(gpkey & PEPC_VK_VOL_DOWN)	Key_Event('c', true); // swim down
    else if (oldgpkey & PEPC_VK_VOL_DOWN) Key_Event('c', false);
    if(gpkey & PEPC_VK_VOL_UP)	Key_Event('d', true); // swim up
    else if (oldgpkey & PEPC_VK_VOL_UP) Key_Event('d', false);
  }
  
  if(gpkey & PEPC_VK_FL && gpkey & PEPC_VK_FR){
	  Key_Event(K_END, true);
	  if(oldgpkey & PEPC_VK_FL)
		  Key_Event(K_ALT, false);
	  if(oldgpkey & PEPC_VK_FR)
		  Key_Event(K_INS, false);
  }
  else if(oldgpkey & PEPC_VK_FL && oldgpkey & PEPC_VK_FR) Key_Event(K_END, false);
  else{
	  if(gpkey & PEPC_VK_FL)	Key_Event(K_ALT, true);
	  else if (oldgpkey & PEPC_VK_FL) Key_Event(K_ALT, false);
	  if(gpkey & PEPC_VK_FR)	Key_Event(K_INS, true);
	  else if (oldgpkey & PEPC_VK_FR) Key_Event(K_INS, false);
  }

 

  oldgpkey=gpkey;

  unsigned char keybuf[128];
  int nread,scancode,pressed,i;

  nread = read(keyboardfd, keybuf, 128);

    for (i = 0; i<nread; ++i) {
      scancode = keybuf[i] & 0x7F;
      if (keybuf[i] & 0x80) {
	pressed = SDL_RELEASED;
      } else {
	pressed = SDL_PRESSED;
      }
      printf("%d %d\n",scancode,pressed);
      if (keymap[scancode]>0) {
	Key_Event(keymap[scancode],pressed);
      }
      //      TranslateKey(scancode, &keysym);
    }


    SDL_Event event;
    int sym, state;
     int modstate;

    while (SDL_PollEvent(&event))
    {
        switch (event.type) {
	  
            case SDL_KEYDOWN:
            case SDL_KEYUP:
                sym = event.key.keysym.sym;
                state = event.key.state;
                modstate = SDL_GetModState();
                switch(sym)
                {
                   case SDLK_DELETE: sym = K_DEL; break;
		   case SDLK_RETURN: sym = K_ENTER; break;
                   case SDLK_BACKSPACE: sym = K_BACKSPACE; break;
                   case SDLK_F1: sym = K_F1; break;
                   case SDLK_F2: sym = K_F2; break;
                   case SDLK_F3: sym = K_F3; break;
                   case SDLK_F4: sym = K_F4; break;
                   case SDLK_F5: sym = K_F5; break;
                   case SDLK_F6: sym = K_F6; break;
                   case SDLK_F7: sym = K_F7; break;
                   case SDLK_F8: sym = K_F8; break;
                   case SDLK_F9: sym = K_F9; break;
                   case SDLK_F10: sym = K_F10; break;
                   case SDLK_F11: sym = K_F11; break;
                   case SDLK_F12: sym = K_F12; break;
                   case SDLK_BREAK:
                   case SDLK_PAUSE: sym = K_PAUSE; break;
                   case SDLK_UP: sym = K_UPARROW; break;
                   case SDLK_DOWN: sym = K_DOWNARROW; break;
                   case SDLK_RIGHT: sym = K_RIGHTARROW; break;
                   case SDLK_LEFT: sym = K_LEFTARROW; break;
                   case SDLK_INSERT: sym = K_INS; break;
                   case SDLK_HOME: sym = K_HOME; break;
                   case SDLK_END: sym = K_END; break;
                   case SDLK_PAGEUP: sym = K_PGUP; break;
                   case SDLK_PAGEDOWN: sym = K_PGDN; break;
                   case SDLK_RSHIFT:
                   case SDLK_LSHIFT: sym = K_SHIFT; break;
                   case SDLK_RCTRL:
                   case SDLK_LCTRL: sym = K_CTRL; break;
                   case SDLK_RALT:
                   case SDLK_LALT: sym = K_ALT; break;
                   case SDLK_KP0: 
                       if(modstate & KMOD_NUM) sym = K_INS; 
                       else sym = SDLK_0;
                       break;
                   case SDLK_KP1:
                       if(modstate & KMOD_NUM) sym = K_END;
                       else sym = SDLK_1;
                       break;
                   case SDLK_KP2:
                       if(modstate & KMOD_NUM) sym = K_DOWNARROW;
                       else sym = SDLK_2;
                       break;
                   case SDLK_KP3:
                       if(modstate & KMOD_NUM) sym = K_PGDN;
                       else sym = SDLK_3;
                       break;
                   case SDLK_KP4:
                       if(modstate & KMOD_NUM) sym = K_LEFTARROW;
                       else sym = SDLK_4;
                       break;
                   case SDLK_KP5: sym = SDLK_5; break;
                   case SDLK_KP6:
                       if(modstate & KMOD_NUM) sym = K_RIGHTARROW;
                       else sym = SDLK_6;
                       break;
                   case SDLK_KP7:
                       if(modstate & KMOD_NUM) sym = K_HOME;
                       else sym = SDLK_7;
                       break;
                   case SDLK_KP8:
                       if(modstate & KMOD_NUM) sym = K_UPARROW;
                       else sym = SDLK_8;
                       break;
                   case SDLK_KP9:
                       if(modstate & KMOD_NUM) sym = K_PGUP;
                       else sym = SDLK_9;
                       break;
                   case SDLK_KP_PERIOD:
                       if(modstate & KMOD_NUM) sym = K_DEL;
                       else sym = SDLK_PERIOD;
                       break;
                   case SDLK_KP_DIVIDE: sym = SDLK_SLASH; break;
                   case SDLK_KP_MULTIPLY: sym = SDLK_ASTERISK; break;
                   case SDLK_KP_MINUS: sym = SDLK_MINUS; break;
                   case SDLK_KP_PLUS: sym = SDLK_PLUS; break;
                   case SDLK_KP_ENTER: sym = SDLK_RETURN; break;
                   case SDLK_KP_EQUALS: sym = SDLK_EQUALS; break;
                }
                // If we're not directly handled and still above 255
                // just force it to 0
                if(sym > 255) sym = 0;
		printf("%d %d\n",sym,state);
                Key_Event(sym, state);
                break;
	  
            case SDL_MOUSEMOTION:
                if ( (event.motion.x != (vid.width/2)) ||
                     (event.motion.y != (vid.height/2)) ) {
                    mouse_x = event.motion.xrel*10;
                    mouse_y = event.motion.yrel*10;
                    if ( (event.motion.x < ((vid.width/2)-(vid.width/4))) ||
                         (event.motion.x > ((vid.width/2)+(vid.width/4))) ||
                         (event.motion.y < ((vid.height/2)-(vid.height/4))) ||
                         (event.motion.y > ((vid.height/2)+(vid.height/4))) ) {
                        SDL_WarpMouse(vid.width/2, vid.height/2);
                    }
                }
                break;

            case SDL_QUIT:
                CL_Disconnect ();
                Host_ShutdownServer(false);        
                Sys_Quit ();
                break;
            default:
                break;
        }
    }

}*/


//char		*argv[MAX_NUM_ARGVS];
//static char *empty_string = "";
//char *lpCmdLine;
//char *origlpCmdLine;

//double	newtime;

int main(int argc,char **argv)
{

        quakeparms_t	parms;
	double		oldtime, time, newtime;
	int t;
	int i;
	
	stickfd = open("/dev/GPIO", O_RDWR | O_NDELAY );
	//if(stickfd < 0){      	
        //printf( "GPIO OPEN FAIL\n"); 
        //return -1;
	//} else {
	//  puts("keyboard emulation started");
	//}

	keyboardfd = open("/dev/tty0", O_RDWR | O_NDELAY );
	if(keyboardfd < 0){      	
        printf( "TTY0 OPEN FAIL\n"); 
        return -1;
	} else {
	  puts("keyboard started");
	}

	doublesize=0;

	//parms.memsize = MINIMUM_MEMORY;
	//parms.memsize = MINIMUM_WIN_MEMORY;
	parms.memsize=10*1024*1024;
	parms.membase = malloc(parms.memsize);
	
	//	mem_dev=open("/dev/mem",O_RDWR);
	//parms.membase=mmap(0, parms.memsize, PROT_READ|PROT_WRITE, MAP_SHARED, mem_dev, 0x3000000);

	//gp2x_dualcore_ram=(unsigned long  *)mmap(0, parms.memsize, PROT_READ|PROT_WRITE, MAP_SHARED, mem_dev, 0x02000000);

	//parms.membase=gp2x_dualcore_ram+0x8000;

	/*
	while (parms.memsize >= MINIMUM_MEMORY) {
	  //		while (parms.memsize >= 0) {
	  //		parms.membase = malloc (parms.memsize);
		parms.membase = malloc(parms.memsize);
		if (!parms.membase) parms.memsize-=300000;
		else break;
	}
	*/
	printf("memory:%d\n",parms.memsize);
	/*			
	char tmpstr[20];
	sprintf(tmpstr,"%d",parms.memsize);
	GpError(tmpstr,69);
	*/

	if (!parms.membase)
	  Sys_Error ("Not enough memory free; check disk space\n");

	parms.basedir = basedir;
	parms.cachedir = cachedir;


	COM_InitArgv (argc, argv);

	parms.argc = com_argc;
	parms.argv = com_argv;

	isDedicated = 0;

    	Sys_Init ();

	Sys_Printf ("Host_Init\n");

	Host_Init (&parms);

	//	GpError("Host_Init", 0);

	//	time = 0.1;
	oldtime = Sys_FloatTime ()-0.1;

	//	gp32_initrendercache();

	puts("here");
    /* main window message loop */

	//		printf("Applying Squidge's MMU hack...\n");
	//	MMUhack(); // Squidge's MMU hack - doesn't do anything as we only use lower 32mb
	//mmuhack();

	while (1)
	{
		newtime = Sys_FloatTime ();
		time = newtime - oldtime;
		if (time>sys_ticrate.value*2)
		  oldtime = newtime;
		else
		  oldtime+=time;

		Host_Frame (time);
	}

	//	GpTimerKill(0);

}

//Dan:
void Sys_HighFPPrecision (void)
{
}

void Sys_LowFPPrecision (void)
{
}
